import os
import uuid
from tempfile import TemporaryDirectory, gettempdir

from octue import exceptions
from octue.templates import copy_template
from ..base import BaseTestCase


class TemplatesTestCase(BaseTestCase):
    """ Test parts of the utils library not otherwise covered by other tests
    """

    def setUp(self):
        super().setUp()
        self.start_path = os.getcwd()
        self.test_id = str(uuid.uuid4())
        self.tmp_dir_name = os.path.join(gettempdir(), "octue-sdk-python", f"test-{self.test_id}")

    def tearDown(self):
        """ Remove data directories manufactured and ensure we end up in the right place
        """
        super().tearDown()
        os.chdir(self.start_path)

    def test_copy_templates_to_directory(self):
        """ Ensures that a known template will copy to a given directory
        """
        with TemporaryDirectory() as tmp_dir:
            copy_template("template-python-fractal", tmp_dir)
            twine_file_name = os.path.join(tmp_dir, "template-python-fractal", "twine.json")
            self.assertTrue(os.path.isfile(twine_file_name))

    def test_copy_templates_to_current_directory_by_default(self):
        """ Ensures that a known template will copy to the current directory by default
        """
        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            copy_template("template-python-fractal")
            self.assertTrue(os.path.isfile(os.path.join(".", "template-python-fractal", "twine.json")))

    def test_copy_template_raises_if_unknown(self):
        """ Ensures that a known template will copy to a given directory
        """
        with TemporaryDirectory() as tmp_dir:
            with self.assertRaises(exceptions.InvalidInputException) as error:
                copy_template("template-which-isnt-there", tmp_dir)
            self.assertIn("Unknown template name 'template-which-isnt-there', try one of", error.exception.args[0])



import os
import sys
import click
import pkg_resources

from octue.definitions import FOLDER_DEFAULTS, MANIFEST_FILENAME, VALUES_FILENAME
from octue.runner import Runner


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--id",
    default=None,
    show_default=True,
    help="Id of the analysis being undertaken. None (for local use) will cause a unique ID to be generated.",
)
@click.option(
    "--skip-checks/--no-skip-checks",
    default=False,
    is_flag=True,
    show_default=True,
    help="Skips the input checking. This can be a timesaver if you already checked "
    "data directories (especially if manifests are large).",
)
@click.option(
    "--log-level",
    default="info",
    type=click.Choice(["debug", "info", "warning", "error"], case_sensitive=False),
    show_default=True,
    help="Log level used for the analysis.",
)
@click.option(
    "--force-reset/--no-force-reset",
    default=True,
    is_flag=True,
    show_default=True,
    help="Forces a reset of analysis cache and outputs [For future use, currently not implemented]",
)
@click.version_option(version=pkg_resources.get_distribution("octue").version)
@click.pass_context
def octue_cli(ctx, id, skip_checks, log_level, force_reset):
    """ Octue CLI, enabling a data service / digital twin to be run like a command line application.

    When acting in CLI mode, results are read from and written to disk (see
    https://octue-python-sdk.readthedocs.io/en/latest/ for how to run your application directly without the CLI).
    Once your application has run, you'll be able to find output values and manifest in your specified --output-dir.
    """
    # TODO Forward command line options to runner via ctx
    ctx.ensure_object(dict)


@octue_cli.command()
@click.option(
    "--app-dir",
    type=click.Path(),
    default=".",
    show_default=True,
    help="Directory containing your source code (app.py)",
)
@click.option(
    "--data-dir",
    type=click.Path(),
    default=".",
    show_default=True,
    help="Location of directories containing configuration values and manifest, input values and manifest, and output "
    "directory.",
)
@click.option(
    "--config-dir",
    type=click.Path(),
    default=None,
    show_default=True,
    help="Directory containing configuration (overrides --data-dir).",
)
@click.option(
    "--input-dir",
    type=click.Path(),
    default=None,
    show_default=True,
    help="Directory containing input (overrides --data-dir).",
)
@click.option(
    "--output-dir",
    type=click.Path(),
    default=None,
    show_default=True,
    help="Directory to write outputs as files (overrides --data-dir).",
)
@click.option(
    "--twine", type=click.Path(), default="twine.json", show_default=True, help="Location of Twine file.",
)
def run(app_dir, data_dir, config_dir, input_dir, output_dir, twine):
    config_dir = config_dir or os.path.join(data_dir, FOLDER_DEFAULTS["configuration"])
    input_dir = input_dir or os.path.join(data_dir, FOLDER_DEFAULTS["input"])
    output_dir = output_dir or os.path.join(data_dir, FOLDER_DEFAULTS["output"])

    runner = Runner(
        twine=twine,
        configuration_values=os.path.join(config_dir, VALUES_FILENAME),
        configuration_manifest=os.path.join(config_dir, MANIFEST_FILENAME),
    )
    analysis = runner.run(
        app_src=app_dir,
        input_values=os.path.join(input_dir, VALUES_FILENAME),
        input_manifest=os.path.join(input_dir, MANIFEST_FILENAME),
        output_manifest_path=os.path.join(output_dir, MANIFEST_FILENAME),
    )
    analysis.finalise(output_dir=output_dir)
    return 0


if __name__ == "__main__":
    args = sys.argv[1:] if len(sys.argv) > 1 else []
    octue_cli(args)



from octue import exceptions
from octue.resources import Datafile, Dataset
from ..base import BaseTestCase


class DatafileTestCase(BaseTestCase):
    def test_instantiates_with_no_args(self):
        """ Ensures a Datafile instantiates using only a path and generates a uuid ID
        """
        Dataset()

    def test_instantiates_with_kwargs(self):
        """ Ensures that keyword arguments can be used to construct the dataset initially
        """
        files = [Datafile(path="path-within-dataset/a_test_file.csv")]
        resource = Dataset(files=files, tags="one two")
        self.assertEqual(len(resource.files), 1)

    def test_append_single_file_to_empty_dataset(self):
        """ Ensures that when a dataset is empty, it can be appended to
        """
        resource = Dataset()
        resource.append(Datafile(path="path-within-dataset/a_test_file.csv"))
        self.assertEqual(len(resource.files), 1)

    def test_append_single_file_to_existing_dataset(self):
        """ Ensures that when a dataset is not empty, it can be appended to
        """
        files = [Datafile(path="path-within-dataset/a_test_file.csv")]
        resource = Dataset(files=files, tags="one two")
        resource.append(Datafile(path="path-within-dataset/a_test_file.csv"))
        self.assertEqual(len(resource.files), 2)

    def test_append_with_datafile_creation_shortcut(self):
        """ Ensures that when a dataset is not empty, it can be appended to
        """
        resource = Dataset()
        resource.append(path="path-within-dataset/a_test_file.csv")
        self.assertEqual(len(resource.files), 1)

    def test_append_multiple_files(self):
        """ Ensures that when a dataset is not empty, it can be appended to
        """
        files = [
            Datafile(path="path-within-dataset/a_test_file.csv"),
            Datafile(path="path-within-dataset/a_test_file.csv"),
        ]
        resource = Dataset()
        resource.append(*files)
        self.assertEqual(len(resource.files), 2)

    def test_cannot_append_non_datafiles(self):
        """ Ensures that exception will be raised if appending a non-datafile object
        """

        class NotADatafile:
            pass

        resource = Dataset()
        with self.assertRaises(exceptions.InvalidInputException) as e:
            resource.append(NotADatafile())

        self.assertIn("must be of class Datafile to append it to a Dataset", e.exception.args[0])

    def test_get_files_catches_single_underscore_mistake(self):
        """ Ensures that if the field name is a single underscore, that gets caught as an error
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/A_Test_file.csv"),
                Datafile(path="path-within-dataset/a_test_file.txt"),
            ]
        )
        with self.assertRaises(exceptions.InvalidInputException) as e:
            resource.get_files("name_icontains", filter_value="Test")

        self.assertIn("Field lookups should be in the form '<field_name>__'<filter_kind>", e.exception.args[0])

    def test_get_files_name_contains(self):
        """ Ensures that get_files works with the name_contains and name_icontains lookups
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/A_Test_file.csv"),
                Datafile(path="path-within-dataset/a_test_file.txt"),
            ]
        )
        files = resource.get_files("name__icontains", filter_value="Test")
        self.assertEqual(2, len(files))
        files = resource.get_files("name__icontains", filter_value="A")
        self.assertEqual(2, len(files))
        files = resource.get_files("name__contains", filter_value="Test")
        self.assertEqual(1, len(files))
        files = resource.get_files("name__icontains", filter_value="test")
        self.assertEqual(2, len(files))
        files = resource.get_files("name__icontains", filter_value="file")
        self.assertEqual(2, len(files))

    def test_get_files_name_with(self):
        """ Ensures that get_files works with the name_endswith and name_startswith lookups
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv"),
                Datafile(path="path-within-dataset/a_your_file.csv"),
            ]
        )
        files = resource.get_files("name__startswith", filter_value="a_my")
        self.assertEqual(1, len(files))
        files = resource.get_files("name__startswith", filter_value="a_your")
        self.assertEqual(1, len(files))
        files = resource.get_files("name__startswith", filter_value="a_")
        self.assertEqual(2, len(files))
        files = resource.get_files("name__startswith", filter_value="b")
        self.assertEqual(0, len(files))
        files = resource.get_files("name__endswith", filter_value="_file.csv")
        self.assertEqual(2, len(files))
        files = resource.get_files("name__endswith", filter_value="r_file.csv")
        self.assertEqual(1, len(files))
        files = resource.get_files("name__endswith", filter_value="y_file.csv")
        self.assertEqual(1, len(files))
        files = resource.get_files("name__endswith", filter_value="other.csv")
        self.assertEqual(0, len(files))

    def test_get_files_by_tag(self):
        """ Ensures that get_files works with tag lookups
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv", tags="one a:2 b:3 all"),
                Datafile(path="path-within-dataset/a_your_file.csv", tags="two a:2 b:3 all"),
                Datafile(path="path-within-dataset/a_your_file.csv", tags="three all"),
            ]
        )
        files = resource.get_files("tag__exact", filter_value="a")
        self.assertEqual(0, len(files))
        files = resource.get_files("tag__exact", filter_value="one")
        self.assertEqual(1, len(files))
        files = resource.get_files("tag__exact", filter_value="all")
        self.assertEqual(3, len(files))
        files = resource.get_files("tag__startswith", filter_value="b")
        self.assertEqual(2, len(files))
        files = resource.get_files("tag__endswith", filter_value="3")
        self.assertEqual(2, len(files))
        files = resource.get_files("tag__contains", filter_value="hre")
        self.assertEqual(1, len(files))

    def test_get_file_by_tag(self):
        """ Ensures that get_files works with tag lookups
        """
        files = [
            Datafile(path="path-within-dataset/a_my_file.csv", tags="one a:2 b:3 all"),
            Datafile(path="path-within-dataset/a_your_file.csv", tags="two a:2 b:3 all"),
            Datafile(path="path-within-dataset/a_your_file.csv", tags="three all"),
        ]

        resource = Dataset(files=files)

        # Check working for single result
        self.assertIs(resource.get_file_by_tag("three"), files[2])

        # Check raises for too many results
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException) as e:
            resource.get_file_by_tag("all")

        self.assertIn("More than one result found", e.exception.args[0])

        # Check raises for no result
        with self.assertRaises(exceptions.UnexpectedNumberOfResultsException) as e:
            resource.get_file_by_tag("billyjeanisnotmylover")

        self.assertIn("No files found with this tag", e.exception.args[0])

    def test_get_files_by_sequence_notnone(self):
        """ Ensures that get_files works with sequence lookups
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv", sequence=0),
                Datafile(path="path-within-dataset/a_your_file.csv", sequence=1),
                Datafile(path="path-within-dataset/a_your_file.csv", sequence=None),
            ]
        )
        files = resource.get_files("sequence__notnone")
        self.assertEqual(2, len(files))

    def test_get_file_sequence(self):
        """ Ensures that get_files works with sequence lookups
        """
        files = [
            Datafile(path="path-within-dataset/a_my_file.csv", sequence=0),
            Datafile(path="path-within-dataset/a_your_file.csv", sequence=1),
            Datafile(path="path-within-dataset/a_your_file.csv", sequence=None),
        ]

        got_files = Dataset(files=files).get_file_sequence("name__endswith", filter_value=".csv", strict=True)
        self.assertEqual(got_files, files[:2])

    def test_get_broken_file_sequence(self):
        """ Ensures that get_files works with sequence lookups
        """
        resource = Dataset(
            files=[
                Datafile(path="path-within-dataset/a_my_file.csv", sequence=2),
                Datafile(path="path-within-dataset/a_your_file.csv", sequence=4),
                Datafile(path="path-within-dataset/a_your_file.csv", sequence=None),
            ]
        )
        with self.assertRaises(exceptions.BrokenSequenceException):
            resource.get_file_sequence("name__endswith", filter_value=".csv", strict=True)

    def test_get_files_name_filters_include_extension(self):
        """ Ensures that filters applied to the name will catch terms in the extension
        """
        files = [
            Datafile(path="path-within-dataset/a_test_file.csv"),
            Datafile(path="path-within-dataset/a_test_file.txt"),
        ]

        self.assertEqual(Dataset(files=files).get_files("name__icontains", filter_value="txt"), [files[1]])

    def test_get_files_name_filters_exclude_path(self):
        """ Ensures that filters applied to the name will not catch terms in the extension
        """
        resource = Dataset(
            files=[
                Datafile(path="first-path-within-dataset/a_test_file.csv"),
                Datafile(path="second-path-within-dataset/a_test_file.txt"),
            ]
        )
        files = resource.get_files("name__icontains", filter_value="second")
        self.assertEqual(0, len(files))



from .base import MixinBase
from .identifiable import Identifiable
from .loggable import Loggable
from .pathable import Pathable
from .serialisable import Serialisable
from .taggable import Taggable


__all__ = "Identifiable", "Loggable", "MixinBase", "Pathable", "Serialisable", "Taggable"



from octue import exceptions
from octue.mixins import MixinBase, Taggable
from octue.mixins.taggable import TagGroup
from ..base import BaseTestCase


class MyTaggable(Taggable, MixinBase):
    pass


class TaggableTestCase(BaseTestCase):
    def test_instantiates(self):
        """ Ensures the class instantiates without arguments
        """
        Taggable()

    def test_instantiates_with_tags(self):
        """ Ensures datafile inherits correctly from the Taggable class and passes arguments through
        """
        tgd = MyTaggable(tags="")
        self.assertEqual("", str(tgd.tags))
        tgd = MyTaggable(tags=None)
        self.assertEqual("", str(tgd.tags))
        tgd = MyTaggable(tags="a b c")
        self.assertEqual("a b c", str(tgd.tags))
        with self.assertRaises(exceptions.InvalidTagException):
            MyTaggable(tags=":a b c")

    def test_instantiates_with_tag_group(self):
        """ Ensures datafile inherits correctly from the Taggable class and passes arguments through
        """
        tgd = MyTaggable(tags="")
        self.assertIsInstance(tgd.tags, TagGroup)
        tgd2 = MyTaggable(tags=tgd.tags)
        self.assertFalse(tgd is tgd2)

    def test_fails_to_instantiates_with_non_iterable(self):
        """ Ensures datafile inherits correctly from the Taggable class and passes arguments through
        """

        class NoIter:
            pass

        with self.assertRaises(exceptions.InvalidTagException) as error:
            MyTaggable(tags=NoIter())

        self.assertIn(
            "Tags must be expressed as a whitespace-delimited string or an iterable of strings", error.exception.args[0]
        )

    def test_reset_tags(self):
        """ Ensures datafile inherits correctly from the Taggable class and passes arguments through
        """
        tgd = MyTaggable(tags="a b")
        tgd.tags = "b c"
        self.assertEqual(str(tgd.tags), "b c")

    def test_valid_tags(self):
        """ Ensures valid tags do not raise an error
        """
        tgd = MyTaggable()
        tgd.add_tags("a-valid-tag")
        tgd.add_tags("a:tag")
        tgd.add_tags("a:-tag")  # <--- yes, this is valid deliberately as it allows people to do negation
        tgd.add_tags("a1829tag")
        tgd.add_tags("1829")
        tgd.add_tags("number:1829")
        tgd.add_tags("multiple:discriminators:used")
        self.assertEqual(
            str(tgd.tags), "a-valid-tag a:tag a:-tag a1829tag 1829 number:1829 multiple:discriminators:used"
        )

    def test_invalid_tags(self):
        """ Ensures invalid tags raise an error
        """

        tgd = MyTaggable()
        with self.assertRaises(exceptions.InvalidTagException):
            tgd.add_tags("-bah")

        with self.assertRaises(exceptions.InvalidTagException):
            tgd.add_tags("humbug:")

        with self.assertRaises(exceptions.InvalidTagException):
            tgd.add_tags("SHOUTY")

        with self.assertRaises(exceptions.InvalidTagException):
            tgd.add_tags(r"back\slashy")

        with self.assertRaises(exceptions.InvalidTagException):
            tgd.add_tags({"not-a": "string"})

    def test_mixture_valid_invalid(self):
        """ Ensures that adding a variety of tags, some of which are invalid, doesn't partially add them to the object
        """
        tgd = MyTaggable()
        tgd.add_tags("first-valid-should-be-added")
        try:
            tgd.add_tags("second-valid-should-not-be-added-because", "-the-third-is-invalid:")

        except exceptions.InvalidTagException:
            pass

        self.assertEqual("first-valid-should-be-added", str(tgd.tags))

    def test_serialises_to_string(self):
        """ Ensures that adding a variety of tags, some of which are invalid, doesn't partially add them to the object
        """
        tgd = Taggable(tags="a b")
        self.assertEqual("a b", tgd.tags.serialise())



import os

from octue import exceptions
from octue.resources import Analysis
from twined import Twine
from ..base import BaseTestCase


class AnalysisTestCase(BaseTestCase):
    def test_instantiate_analysis(self):
        """ Ensures that the base analysis class can be instantiated
        """
        # from octue import runner  # <-- instantiated in the library, not here
        analysis = Analysis(twine="{}")
        self.assertEqual(analysis.__class__.__name__, "Analysis")

    def test_instantiate_analysis_with_twine(self):
        """ Ensures that the base analysis class can be instantiated
        """
        analysis = Analysis(twine=Twine(source="{}"))
        self.assertEqual(analysis.__class__.__name__, "Analysis")

    def test_protected_setter(self):
        """ Ensures that protected attributes can't be set
        """
        analysis = Analysis(twine="{}")
        with self.assertRaises(exceptions.ProtectedAttributeException) as error:
            analysis.configuration_values = {}

        self.assertIn("You cannot set configuration_values on an instantiated Analysis", error.exception.args[0])

    def test_protected_getter(self):
        """ Ensures that protected attributes can't be set
        """
        analysis = Analysis(
            twine=str(os.path.join(self.data_path, "twines", "valid_schema_twine.json")),
            configuration_values={"n_iterations": 5},
            input_values={"height": 5},
            output_values={},
        )
        cfg = analysis.configuration_values
        self.assertIn("n_iterations", cfg.keys())



import os

from octue.exceptions import InvalidInputException


class Pathable:
    """ Mixin class to enable resources to get their path location from an owner.

    For example, datasets can get their path from the Manifest they belong to.

    """

    def __init__(self, *args, path=None, path_from=None, base_from=None, **kwargs):
        """ Constructor for pathable mixin
        """
        super().__init__(*args, **kwargs)

        if (path_from is not None) and not isinstance(path_from, Pathable):
            raise InvalidInputException(
                "paths_from argument must be an instance of an object inheriting from Pathable() mixin"
            )

        if (base_from is not None) and not isinstance(base_from, Pathable):
            raise InvalidInputException(
                "base_from argument must be an instance of an object inheriting from Pathable() mixin"
            )

        self._path_from = path_from
        self._base_from = base_from
        self._path_is_absolute = False
        self.path = path

    @property
    def _base_path(self):
        """ Gets the absolute path of the base_from object, from which any relative paths are constructed
        :return:
        :rtype:
        """
        if self._base_from:
            return self._base_from.absolute_path

        return os.getcwd()

    @property
    def _path_prefix(self):
        """ Gets the path prefix (this is the absolute_path of the owner path_from object).
        Defaults to the current working directory
        """
        if self._path_from:
            return self._path_from.absolute_path

        if self._path_is_absolute:
            return ""

        return os.getcwd()

    @property
    def absolute_path(self):
        """ The absolute path of this resource
        """
        return os.path.normpath(os.path.join(self._path_prefix, self._path))

    @property
    def relative_path(self):
        """ The path of this resource relative to its base path
        """
        return os.path.relpath(self.absolute_path, self._base_path)

    @property
    def path(self):
        """ The path of this resource
        """
        return self._path

    @path.setter
    def path(self, value):
        """ Set the path of this resource.

        :param value: Path of the resource. If the resource was instantiated with a `path_from` object, this path must
        be relative. Otherwise, absolute paths are acceptable.
        :type value: Union[str, path-like]
        """

        value = os.path.normpath(value or ".")

        path_is_absolute = value == os.path.abspath(value)

        if path_is_absolute and self._path_from is not None:
            raise InvalidInputException(
                f"You cannot an absolute path on a pathable instantiated with 'path_from'. Set a path relative to the path_from object ({self._path_from})"
            )

        self._path_is_absolute = path_is_absolute
        self._path = value



import logging

from octue.exceptions import InvalidInputException, InvalidManifestException
from octue.mixins import Identifiable, Loggable, Pathable, Serialisable
from .dataset import Dataset


module_logger = logging.getLogger(__name__)


class Manifest(Pathable, Serialisable, Loggable, Identifiable):
    """ A representation of a manifest, which can contain multiple datasets
    This is used to manage all files coming into (or leaving), a data service for an analysis at the
    configuration, input or output stage.
    """

    def __init__(self, id=None, logger=None, path=None, path_from=None, base_from=None, **kwargs):
        """ Construct a Manifest
        """
        super().__init__(id=id, logger=logger, path=path, path_from=path_from, base_from=base_from)

        # TODO The decoders aren't being used; utils.decoders.OctueJSONDecoder should be used in twined
        #  so that resources get automatically instantiated.
        #  Add a proper `decoder` argument  to the load_json utility in twined so that datasets, datafiles and manifests
        #  get initialised properly, then tidy up this hackjob. Also need to allow Pathables to update ownership
        #  (because decoders work from the bottom of the tree upwards, not top-down)

        datasets = kwargs.pop("datasets", list())
        self.keys = kwargs.pop("keys", dict())

        # TODO we need to add keys to the manifest file schema in twined so that we know what dataset(s) map to what keys
        #  In the meantime, we enforce at this level that keys will match
        n_keys = len(self.keys.keys())
        n_datasets = len(datasets)
        if n_keys != n_datasets:
            raise InvalidManifestException(
                f"Manifest instantiated with {n_keys} keys, and {n_datasets} datasets... keys must match datasets!"
            )

        # Sort the keys by the dataset index so we have a list of keys in the same order as the dataset list.
        # We'll use this to name the dataset folders
        key_list = [k for k, v in sorted(self.keys.items(), key=lambda item: item[1])]

        # Instantiate the datasets if not already done
        self.datasets = []
        for key, ds in zip(key_list, datasets):
            if isinstance(ds, Dataset):
                self.datasets.append(ds)
            else:
                self.datasets.append(Dataset(**ds, path=key, path_from=self))

        # Instantiate the rest of everything!
        self.__dict__.update(**kwargs)

    def get_dataset(self, key):
        """ Gets a dataset by its key name (as defined in the twine)
        :return: Dataset selected by its key
        :rtype: Dataset
        """
        idx = self.keys.get(key, None)
        if idx is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset '{key}' from Manifest. Allowable keys are: {list(self.keys.keys())}"
            )

        return self.datasets[idx]

    def prepare(self, data):
        """ Prepare new manifest from a manifest_spec
        """
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for idx, dataset_spec in enumerate(data):

            self.keys[dataset_spec["key"]] = idx
            # TODO generate a unique name based on the filter key, tag datasets so that the tag filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets.append(Dataset(logger=self.logger, path_from=self, path=dataset_spec["key"]))

        return self



import os
import shutil
import sys
import uuid

from octue import Runner
from ..base import BaseTestCase


class TemplateAppsTestCase(BaseTestCase):
    """ Test case that runs analyses using apps in the templates, to ensure all the examples work.
    """

    def setUp(self):
        super().setUp()
        self.start_path = os.getcwd()

        # Initialise so these variables are assigned on the instance
        self.template_data_path = None
        self.template_twine = None
        self.template_path = None
        self.app_test_path = None
        self.teardown_templates = []

        def set_template(template):
            """ Sets up the working directory and data paths to run one of the provided templates
            """
            self.template_path = os.path.join(self.templates_path, template)
            self.template_twine = os.path.join(self.templates_path, template, "twine.json")

            # Duplicate the template's data/ directory to a test-specific replica
            self.app_test_path = os.path.join(self.data_path, str(uuid.uuid4()))
            shutil.copytree(self.template_path, self.app_test_path)

            # Add this template to the list to remove in teardown
            self.teardown_templates.append(self.app_test_path)
            sys.path.insert(0, self.app_test_path)

            # Run from within the app folder context
            os.chdir(self.app_test_path)

        self.set_template = set_template

    def tearDown(self):
        """ Remove data directories manufactured
        """
        super().tearDown()
        os.chdir(self.start_path)
        for path in self.teardown_templates:
            sys.path.remove(path)
            shutil.rmtree(path)

    def test_fractal_configuration(self):
        """ Ensures fractal app can be configured with its default configuration
        """
        self.set_template("template-python-fractal")
        runner = Runner(
            twine=self.template_twine,
            configuration_values=os.path.join("data", "configuration", "configuration_values.json"),
        )
        analysis = runner.run(
            app_src=self.template_path, output_manifest_path=os.path.join("data", "output", "manifest.json")
        )
        analysis.finalise(output_dir=os.path.join("data", "output"))

    def test_using_manifests(self):
        """ Ensures using-manifests app works correctly
        """
        self.set_template("template-using-manifests")
        runner = Runner(
            twine=self.template_twine, configuration_values=os.path.join("data", "configuration", "values.json"),
        )
        analysis = runner.run(
            app_src=self.template_path,
            input_manifest=os.path.join("data", "input", "manifest.json"),
            output_manifest_path=os.path.join("data", "output", "manifest.json"),
        )
        analysis.finalise(output_dir=os.path.join("data", "output"))
        self.assertTrue(os.path.isfile(os.path.join("data", "output", "cleaned_met_mast_data", "cleaned.csv")))



import os
import uuid

from octue import exceptions
from octue.mixins import MixinBase, Pathable
from octue.resources import Datafile
from ..base import BaseTestCase


class MyPathable(Pathable, MixinBase):
    pass


class DatafileTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.path_from = MyPathable(path=os.path.join(self.data_path, "basic_files", "configuration", "test-dataset"))
        self.path = os.path.join("path-within-dataset", "a_test_file.csv")

    def create_valid_datafile(self):
        return Datafile(path_from=self.path_from, base_from=self.path_from, path=self.path, skip_checks=False)

    def test_instantiates(self):
        """ Ensures a Datafile instantiates using only a path and generates a uuid ID
        """
        df = Datafile(path="a_path")
        self.assertTrue(isinstance(df.id, str))
        self.assertEqual(type(uuid.UUID(df.id)), uuid.UUID)
        self.assertIsNone(df.sequence)
        self.assertEqual(0, df.cluster)

    def test_path_argument_required(self):
        """ Ensures instantiation without a path will fail
        """
        with self.assertRaises(exceptions.InvalidInputException) as error:
            Datafile()

        self.assertIn("You must supply a valid 'path' for a Datafile", error.exception.args[0])

    def test_checks_fail_when_file_doesnt_exist(self):
        path = "not_a_real_file.csv"
        with self.assertRaises(exceptions.FileNotFoundException) as error:
            Datafile(path=path, skip_checks=False)
        self.assertIn("No file found at", error.exception.args[0])

    def test_conflicting_extension_fails_check(self):
        with self.assertRaises(exceptions.InvalidInputException) as error:
            Datafile(path_from=self.path_from, path=self.path, skip_checks=False, extension="notcsv")

        self.assertIn("Extension provided (notcsv) does not match file extension", error.exception.args[0])

    def test_file_attributes_accessible(self):
        """ Ensures that its possible to set the sequence, cluster and timestamp
        """
        df = self.create_valid_datafile()
        self.assertIsInstance(df.size_bytes, int)
        self.assertGreaterEqual(df.last_modified, 1598200190.5771205)
        self.assertEqual("a_test_file.csv", df.name)

        df.sequence = 2
        df.cluster = 0
        df.posix_timestamp = 0

    def test_cannot_set_calculated_file_attributes(self):
        """ Ensures that calculated attributes cannot be set
        """
        df = self.create_valid_datafile()

        with self.assertRaises(AttributeError):
            df.size_bytes = 1

        with self.assertRaises(AttributeError):
            df.last_modified = 1000000000.5771205

    def test_serialisable(self):
        """ Ensures a datafile can serialise to json format
        """
        df = self.create_valid_datafile()
        df_dict = df.serialise()

        for k in df_dict.keys():
            self.assertFalse(k.startswith("_"))

        for k in (
            "cluster",
            "extension",
            "id",
            "last_modified",
            "name",
            "path",
            "posix_timestamp",
            "sequence",
            "size_bytes",
            "tags",
            "sha_256",
        ):
            self.assertIn(k, df_dict.keys())



import hashlib
import logging
import os
import time

from octue.exceptions import FileNotFoundException, InvalidInputException
from octue.mixins import Identifiable, Loggable, Pathable, Serialisable, Taggable
from octue.utils import isfile


module_logger = logging.getLogger(__name__)


class Datafile(Taggable, Serialisable, Pathable, Loggable, Identifiable):
    """ Class for representing data files on the Octue system

    Files in a manifest look like this:

        {
          "path": "folder/subfolder/file_1.csv",
          "cluster": 0,
          "sequence": 0,
          "extension": "csv",
          "tags": "",
          "posix_timestamp": 0,
          "id": "abff07bc-7c19-4ed5-be6d-a6546eae8e86",
          "last_modified": "2019-02-28T22:40:30.533005Z",
          "size_bytes": 59684813,
          "sha-512/256": "somesha"
        },

    :parameter path_from: The root Pathable object (typically a Dataset) that this Datafile's path is relative to.
    :type path_from: Pathable

    :parameter base_from: A Pathable object, which in most circumstances is the same as the path_from object, upon which
    the `relative_path` property is based (if not given, `relative_path` is relative to current working directory

    :parameter path: The path of this file, which may include folders or subfolders, within the dataset. If no path_from
    parameter is set, then absolute paths are acceptable, otherwise relative paths are required.
    :type path: Union[str, path-like]

    :parameter logger: A logger instance to which operations with this datafile will be logged. Defaults to the module logger.
    :type logger: logging.Logger

    :parameter id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
    :type id: str

    :parameter cluster: The cluster of files, within a dataset, to which this belongs (default 0)
    :type cluster: int

    :parameter sequence: A sequence number of this file within its cluster (if sequences are appropriate)
    :type sequence: int

    :parameter tags: Space-separated string of tags relevant to this file
    :type tags: str

    :parameter posix_timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it
    was created but could relate to a relevant time point for the data
    :type posix_timestamp: number
    """

    _exclude_serialise_fields = ("logger", "open")

    def __init__(
        self,
        id=None,
        logger=None,
        path=None,
        path_from=None,
        base_from=None,
        cluster=0,
        sequence=None,
        tags=None,
        posix_timestamp=None,
        skip_checks=True,
        **kwargs,
    ):
        """ Construct a datafile
        """
        super().__init__(id=id, logger=logger, tags=tags, path=path, path_from=path_from, base_from=base_from)

        self.cluster = cluster

        self.sequence = sequence
        self.posix_timestamp = posix_timestamp or time.time()

        if path is None:
            raise InvalidInputException("You must supply a valid 'path' for a Datafile")

        # Set up the file extension or get it from the file path if none passed
        self.extension = self._get_extension_from_path()

        # Run integrity checks on the file
        if not skip_checks:
            self.check(**kwargs)

    def _get_extension_from_path(self, path=None):
        """ Gets extension of a file, either from a provided file path or from self.path field
        """
        path = path or self.path
        return os.path.splitext(path)[-1].strip(".")

    def _get_sha_256(self):
        """ Calculate the SHA256 hash string of the file
        """
        sha256_hash = hashlib.sha256()
        with open(self.absolute_path, "rb") as f:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)

        return sha256_hash.hexdigest()

    @property
    def name(self):
        return str(os.path.split(self.path)[-1])

    @property
    def last_modified(self):
        return os.path.getmtime(self.absolute_path)

    @property
    def size_bytes(self):
        return os.path.getsize(self.absolute_path)

    @property
    def sha_256(self):
        return self._get_sha_256()

    def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
        """ Check file presence and integrity
        """
        # TODO Check consistency of size_bytes input against self.size_bytes property for a file if we have one
        # TODO Check consistency of sha against file contents if we have a file
        # TODO Check consistency of last_modified date

        if (extension is not None) and not self.path.endswith(extension):
            raise InvalidInputException(
                f"Extension provided ({extension}) does not match file extension (from {self.path}). Pass extension=None to set extension from filename automatically."
            )

        if not self.exists():
            raise FileNotFoundException(f"No file found at {self.absolute_path}")

    def exists(self):
        """ Returns true if the datafile exists on the current system, false otherwise
        """
        return isfile(self.absolute_path)

    @property
    def open(self):
        """ Context manager to handle the opening and closing of a Datafile.

        If opened in write mode, the manager will attempt to determine if the folder path exists and, if not, will
        create the folder structure required to write the file.

        Use it like:
        ```
        my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
        with my_datafile.open('w') as fp:
            fp.write("{}")
        ```
        This is equivalent to the standard python:
        ```
        my_datafile = Datafile(path='subfolder/subsubfolder/my_datafile.json)
        os.makedirs(os.path.split(my_datafile.absolute_path)[0], exist_ok=True)
        with open(my_datafile.absolute_path, 'w') as fp:
            fp.write("{}")
        ```
        """

        absolute_path = self.absolute_path

        class DataFileContextManager:
            def __init__(obj, mode="r", **kwargs):
                obj.mode = mode
                obj.kwargs = kwargs
                obj.absolute_path = absolute_path
                if "w" in obj.mode:
                    os.makedirs(os.path.split(obj.absolute_path)[0], exist_ok=True)

            def __enter__(obj):
                obj.fp = open(obj.absolute_path, obj.mode, **obj.kwargs)
                return obj.fp

            def __exit__(obj, *args):
                if obj.fp is not None:
                    obj.fp.close()

        return DataFileContextManager
